"""Holder

A class to represent a owner of subjects.

https://github.com/rochars/trade
License: MIT

Copyright (c) 2015-2018 Rafael da Silva Rocha

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""


class Holder(object):
    """A class to represent a owner of subjects.

    A Holder is created with a initial state regarding the subjects it
    owns. The state of Holder is then updated by Occurrence objects by
    calling Holder.trade(occurrence).

    trade apps should define the Holder.state signature according to
    their needs.

    Classes that inherit from Occurrence may add and manipulate new
    items on the Holder.state dictionary.

    Attributes:
        state: A dict representing the state of the Holder.
            The example implementation uses the following signature:
            {
                "SUBJECT SYMBOL": {
                    "quantity": 0,
                    "value": 0
                },
                ...
            }
    """

    def __init__(self, state=None):
        self.state = {}
        if state:
            self.state = state

    def trade(self, occurrence):
        """Update the state of the Holder with an occurrence."""
        occurrence.update_holder(self)



"""Context

A class to group occurrences and contextualize them.

https://github.com/rochars/trade
License: MIT

Copyright (c) 2015-2018 Rafael da Silva Rocha

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

from __future__ import absolute_import
from __future__ import division

import copy


class Context(object):
    """A class to group occurrences and contextualize them.

    To contextualize a group of occurrences could be:
    - To identify daytrades and separate them into different occurrences
      for taxation purposes
    - Group occurrences with the same subject into one occurrence
    - Apply costs and other deductions to the value to the occurrences
    - Calculate taxes based on the occurrences in the context

    Objects from this class are created with both a list of occurrences and a
    list of functions (called "context tasks").

    The context tasks are run by calling Context.apply(). Context tasks
    run in the order they are listed.

    Every context task function receives one argument: the context itself. It
    is then free to manipulate the raw occurrence data stored in the
    Context.occurrences list.

    The contextualized occurrence data is then stored in the
    Context.contextualized list.

    The occurrences in Context.contextualized should be informed to a Holder
    as regular occurrences.

    The original Occurrence list is not modified.
    """

    def __init__(self, occurrences, tasks):
        self.occurrences = occurrences
        self.tasks = tasks
        self.contextualized = []

    def apply(self):
        """Apply the rules of the context to its occurrences.

        This method executes all the functions defined in
        self.tasks in the order they are listed.

        Every function that acts as a context task receives the
        Context object itself as its only argument.

        The contextualized occurrences are then stored in
        Context.contextualized.

        The original Occurrence instances are not modified.
        """
        raw_operations = copy.deepcopy(self.occurrences)
        for task in self.tasks:
            task(self)
        self.occurrences = raw_operations



"""Occurrence

A class to represent occurrences with subjects.

https://github.com/rochars/trade
License: MIT

Copyright (c) 2015-2018 Rafael da Silva Rocha

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

from __future__ import absolute_import
from __future__ import division


class Occurrence(object):
    """A class to represent occurrences with subjects.

    Attributes:
        subject: A Subject instance.
        date: A string to represent the moment of the occurrence.
        details: A dictionary with the occurrence details.
    """

    def __init__(self, subject, date, details):
        self.subject = subject
        self.date = date
        self.details = details

    def update_holder(self, holder):
        """Udpate the Holder state according to the occurrence.

        This implementation is a example of how a Occurrence object
        can update the Holder state; this method should be overriden
        by classes that inherit from the Occurrence class.

        This sample implementation simply update the quantity and the average
        price of the Subject in the Holder's possession every time objects
        from this class are passed to Holder.trade().

        This sample implementation considers the following signature for
        the Holder.state dictionary:
            {
                "SUBJECT SYMBOL": {
                    "quantity": 0,
                    "value": 0
                },
                ...
            }
        """

        subject_symbol = self.subject.symbol

        # If the Holder already have a state regarding this Subject,
        # update that state
        if subject_symbol in holder.state:

            # If the Holder have zero units of this subject, the average
            # value paid/received for the subject is the value of the trade itself
            if not holder.state[subject_symbol]['quantity']:
                holder.state[subject_symbol]['value'] = self.details['value']

            # If the Holder owns units of this subject then the average value
            # paid/received for the subject may need to be updated with
            # this occurrence details
            elif same_sign(
                    holder.state[subject_symbol]['quantity'],
                    self.details['quantity']):
                holder.state[subject_symbol]['value'] = average_price(
                    holder.state[subject_symbol]['quantity'],
                    holder.state[subject_symbol]['value'],
                    self.details['quantity'],
                    self.details['value']
                )

            # Update the quantity of the subject in the Holder's posession
            holder.state[subject_symbol]['quantity'] += self.details['quantity']

        # If the Holder don't have a state with this occurrence's Subject,
        # then register this occurrence as the first state of the Subject
        # in the Holder's possession
        else:
            holder.state[subject_symbol] = {
                'quantity': self.details['quantity'],
                'value': self.details['value']
            }

        # If the Holder knows about this Subject but don't have any unit
        # of it, the paid value of the subject in the Holder state should
        # be zero.
        if not holder.state[subject_symbol]['quantity']:
            holder.state[subject_symbol]['value'] = 0

def average_price(quantity_1, price_1, quantity_2, price_2):
    """Calculates the average price between two asset states."""
    return (quantity_1 * price_1 + quantity_2 * price_2) / \
            (quantity_1 + quantity_2)

def same_sign(number_1, number_2):
    """Checks if two numbers have the same sign."""
    return (number_1 >= 0) ^ (number_2 < 0)
